import numpy as np
import math

"""
Weights are used to model situations, where categories are represented as (at least) ordinal data.
Using this approach, the agreement between raters is not binary, but it differs depending on the
weights between chosen categories.

For example, while modeling ratio categories with possible values: `{1, 2, 3, 4, 5}`, values 
`5` and `4` are will cause higher agreement than `5` and `1`.

This implementation assumes all categories are evenly distributed. That means the distance between
each category is exactly the same. You can change that by implementing a custom weight kernel.

There is no formal rule that can be used for deciding which set weights should be used
in a particular inter-rater reliability study. The researcher needs to determine how
much each type of disagreement (or partial agreement) should impact the agreement
coefficient, and select the set of weights accordingly.
"""


def identity_kernel(x: int, y: int, q: int) -> float:
    """
    The default kernel, which don't introduce any weights between categories.
    Two values get any positive agreement only if they have exactly the same category.
    """
    return float(x == y)


def linear_kernel(x: int, y: int, q: int) -> float:
    """
    The most basic weight kernel.
    The weights' values depend on whether the values are of alphabetic or numeric type.
    The linear weights are generally smaller than the quadratic weights.
    """
    return 1 - abs(x - y) / (q - 1)


def quadratic_kernel(x: int, y: int, q: int) -> float:
    return 1 - (x - y)**2 / (q - 1)**2


def radical_kernel(x: int, y: int, q: int) -> float:
    """
    This kernel can be used if you find quadratic and linear weights too large.
    """
    return 1 - np.sqrt(abs(x - y)) / np.sqrt((q - 1))


def radio_kernel(x: int, y: int, q: int) -> float:
    """
    This kernel can be used with data of ratio type.
    Ratio weights evaluate the differences between scores relative to their magnitudes.
    """
    _x, _y = x + 1, y + 1
    return 1 - ((_x - _y)/(_x + _y))**2 / ((q - 1)/(q + 1))**2


def ordinal_kernel(x: int, y: int, q: int) -> float:
    """
    This kernel can be used for non-numeric ordinal values.
    The use of ordinal weights requires the categories to be at least of ordinal type.
    So you should be able to rank all categories from the smallest to the largest.
    """
    def m(k, l):
        return math.comb(max(k, l) - min(k, l) + 1, 2)

    if x == y:
        return 1.0
    else:
        return 1 - m(x, y) / m(1, q)


def circular_kernel(x: int, y: int, q: int) -> float:
    """
    These weights would be recommendedif the rating represents the magnitude
    of an angle expressed in degrees or in radians.
    """
    return 1 - (np.sin(np.pi * (x - y) / q) / np.sin(np.pi * (int(q / 2)) / q))**2


def bipolar_kernel(x: int, y: int, q: int) -> float:
    """
    The bipolar weights, which behave like ratio weights at the center of the scale and
    like quadratic weights towards the ends.
    """
    if x == y:
        return 1.0
    else:
        _x, _y = x + 1, y + 1
        return 1 - (_x - _y)**2 / ((_x + _y - 2) * (2 * q - _x - _y))


def compute_weights(q: int, kernel):
    return np.array([
        [
            kernel(x, y, q) for x in range(q)
        ] for y in range(q)
    ])



import numpy as np
from typing import Tuple


def _encode_one_hot(values: np.ndarray, columns: np.ndarray) -> np.ndarray:
    """Used to create a one-hot encoding table."""
    _temp = np.zeros((len(values), len(columns)))
    _temp[range(len(values)), values] = 1
    return _temp


def _get_inverted_index(x: np.ndarray, values: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Used to create a inverted index for non-dense array."""
    cols = values[np.argsort(values)]
    return cols, np.searchsorted(cols, x)


def pivot_table_frequency(rows: np.ndarray, columns: np.ndarray, values: np.ndarray = None) -> np.ndarray:
    """
    Used to create a pivot table containing a frequency of each column value for each unique row value.
    Rows can be treated as keys and columns as values. This function creates an array, where each row
    contains a frequency of all possible values for one key.


    :param rows: specifies rows of the table
    :param columns: specifies columns of the table
    :param values: used to provide a predefined list of values for columns.
    Can be used when some of the options don't appear in the columns data.
    :return: a pivot table containing frequencies
    """
    rows, row_pos = np.unique(rows, return_inverse=True)
    if values is None:
        cols, col_pos = np.unique(columns, return_inverse=True)
    else:
        cols, col_pos = _get_inverted_index(columns, values)
    return _encode_one_hot(row_pos, rows).T.dot(_encode_one_hot(col_pos, cols))




import numpy as np
from agreement.utils.kernels import (
    identity_kernel, compute_weights
)


def _compute_observed_agreement(answers_matrix, weights_kernel=identity_kernel):
    N, q = answers_matrix.shape
    n = answers_matrix.sum(axis=1)

    w = compute_weights(q, weights_kernel)

    r_star = answers_matrix.dot(w)

    po = np.average((answers_matrix * (r_star-1)).sum(axis=1) / (n * (n-1)))
    return po, w


def observed_agreement(answers_matrix, weights_kernel=identity_kernel):
    answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]
    po, _ = _compute_observed_agreement(answers_matrix, weights_kernel)
    return po


def s_score(answers_matrix, weights_kernel=identity_kernel):
    answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]
    N, q = answers_matrix.shape

    po, w = _compute_observed_agreement(answers_matrix, weights_kernel)

    pc = w.sum() / q**2

    S = (po - pc) / (1 - pc)
    return S


def cohens_kappa(answers_matrix, users_matrix, weights_kernel=identity_kernel):
    answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]

    N, q = answers_matrix.shape

    po, w = _compute_observed_agreement(answers_matrix, weights_kernel)

    p = users_matrix / users_matrix.sum(axis=1, keepdims=True)

    r, q = p.shape

    pbar = p.sum(axis=0) / r

    rpbar = r * pbar.reshape(1, q).T * pbar.reshape(1, q)
    pg = p.T.dot(p)
    s2 = (pg - rpbar) / (r - 1)

    pbarplus = pbar.reshape(1, q).T * pbar.reshape(1, q)
    pc = (w * (pbarplus - s2/r)).sum()

    k = (po - pc) / (1 - pc)
    return k


def gwets_gamma(answers_matrix, weights_kernel=identity_kernel):
    N, q = answers_matrix.shape

    _answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]
    po, w = _compute_observed_agreement(_answers_matrix, weights_kernel)

    Tw = w.sum()

    pi = np.average(answers_matrix / answers_matrix.sum(axis=1, keepdims=True), axis=0)

    pc = (pi*(1 - pi)).sum() * Tw / (q*(q-1))

    gamma = (po - pc) / (1 - pc)
    return gamma



def krippendorffs_alpha(answers_matrix, weights_kernel=identity_kernel):
    _answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]

    N, q = _answers_matrix.shape
    n = _answers_matrix.sum(axis=1)

    w = compute_weights(q, weights_kernel)

    r_star = _answers_matrix.dot(w)

    rdash = _answers_matrix.sum(axis=1).sum() / N
    epsilon = 1 / (N * rdash)

    po2 = (_answers_matrix * (r_star-1) / (rdash * (n - 1))[:,np.newaxis]).sum(axis=1).mean()

    po = po2 * (1 - epsilon) + epsilon

    pc = ((_answers_matrix.mean(axis=0) / rdash)**2).sum()

    alpha = (po - pc) / (1 - pc)
    return alpha


def scotts_pi(answers_matrix, weights_kernel=identity_kernel):
    answers_matrix = answers_matrix[answers_matrix.sum(axis=1) > 1]
    N, q = answers_matrix.shape

    po, w = _compute_observed_agreement(answers_matrix, weights_kernel)

    pik = np.mean(answers_matrix / answers_matrix.sum(axis=1, keepdims=True), axis=0)

    pc = (w * pik.reshape(1, q).T * pik.reshape(1, q)).sum()

    pi = (po - pc) / (1 - pc)
    return pi

